/*
 Processing User Interface Elements
 Compatible with Processing.js (some default font differences apply, and vaires with browser)
 
 Released under The MIT License
 
 Copyright (c) 2011 Mick West
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 
 */
 
 /*
 
 Example usage:

// Variables, these can be globals for simplicity. 

// At a minimum you need one manager object. This handles updating things, and drawing them.
MUIElementManager MUIMan;

// Variables for sliders
MUIVSlider speedSlider,windSlider,headingSlider,turnrateSlider;

// Variables for Check Boxes
MUICheckBox pauseCheckBox, dirCheckBox, fig8CheckBox;

// Varaible for dragable boxes
MUIDragable P1Drag, P2Drag, CenterDrag;

// INITIALIZATION

// This would be somethwere in your setup code. You need to create the Element manager

  // MUISlider Manager
  MUIMan = new MUIElementManager();

// Then add elements
// Vslider are vertical sliders. 
//  parameters are (float x_, float y_, float w_, float h_, float start_, float end_, float initial_, float _resolution, String _lable)

  speedSlider =      MUIMan.addVSlider(450,400,10,200,100,1000,478,1,"Plane\nSpeed\nMPH");
  windSlider =       MUIMan.addVSlider(500,400,10,200,0,100,44,1,"Wind\nSpeed\nMPH");
  headingSlider =    MUIMan.addVSlider(550,400,10,200,0,360,59,1,"Wind\nHeading\nDeg");
  turnrateSlider =   MUIMan.addVSlider(600,400,10,200,60,1200,360,10,"Turn\nrate\nsecs");
  
  pauseCheckBox = MUIMan.addCheckBox(450,620,10,10,0,"Pause");
  dirCheckBox =   MUIMan.addCheckBox(520,620,10,10,0,"Clockwise");
  fig8CheckBox = MUIMan.addCheckBox(380,620,10,10,0,"Fig8");
  
  P1Drag = MUIMan.addDragable(50*miles_to_meters/meters_per_pixel,24*miles_to_meters/meters_per_pixel,10,10,0,"P1");
  P2Drag = MUIMan.addDragable(73*miles_to_meters/meters_per_pixel,25*miles_to_meters/meters_per_pixel,10,10,0,"P2");
  
  CenterDrag = MUIMan.addDragable((P1Drag.x+P2Drag.x)/2,(P1Drag.y+P2Drag.y)/2,10,10,0,"");
 
 Then you just need to call the "update" and "display" functions
 
  MUIMan.update();
  MUIMan.display();
  
  The UI elements will be displayed, and you can then access their state (for sliders and check boxes) and position (for draggable points).
   
 */
 

class MUIElement
{
  float x,y;                    // position of the center of the element in screen space  
  float value;                  // value of the slider calculated 
  float w,h;                    // width and height (not currently the actual width and height for sliders, the border is added) 
  String  lable;                // lable below the slider. Can have line feeds, like  "max\nspeed\nin MPH"
  boolean alive = true;         // flag if alive. setting it to false will disable it, and it will be pruned by the slider manager
  float border_x = 0;          // border to left and right of the slider
  float border_y = 0;          // border above and below
  
  boolean visible = true;     // flag if it is displayable


  void display() {
  };
  void update() {
  };

  boolean dead()
  {
    return !alive;
  }

  void die()
  {
    alive = false;
  }

  void hide()
  {
    visible = false;
  }
  
  void show()
  {
    visible = true;
  }
  
  // handle a mouse click
  // returns true if we've handled it, and don't want to propogate
  boolean  handleClick(float x, float y) {  return false;  }
  boolean  handleRelease(float x, float y) {  return false;  }
  boolean  handleMove(float x, float y) {  return false;  }
  
}

class MUIVSlider extends MUIElement
{
  float yp;                     // y position of the slider bar in screen space
  float start, end;             // start (slider all the way down) and end (slider all the way up) values
  float resolution;             // the smallest step that the slider value is output in.

  float button_height = 10;     // height of the slider button, w is used as the width of the button 
  boolean dragging = false;
 
  // Constructor, generally this will be called by a slider manager
  MUIVSlider(float x_, float y_, float w_, float h_, float start_, float end_, float initial_, float _resolution, String _lable)
  {
    x = x_;
    y = y_;
    w = w_;
    h = h_;

    resolution = _resolution;

    start=start_;
    end=end_;

    lable = _lable;

    value = (initial_ - start) / (end - start);
    yp = y-h/2+h*(1-value);
    
    border_x = 10;
    border_y = 10;
    
  }


  void display()
  {
    rectMode(CENTER);

    // Background rectangle
    fill(40);
    noStroke();
    rect(x,y,w+border_x,h+button_height+border_y);

    stroke(255);
    strokeWeight(4);
    line(x,y-h/2,x,y+h/2); // The center line

    strokeWeight(1);
    fill(128); 
    rect(x,yp,w,button_height);
    strokeWeight(3);
    line(x+w/2,yp,x-w/2,yp);

    pushMatrix();    
    float  s = 1;
    scale (s);
    fill(0,255,0,255);
    textAlign(CENTER);
    // processing.js does not support the extra width and height parameters
    // so we just use the x,y version, which works fine with a little tweaking.
    text(""+getValue(),x/s,(y-h/2-border_y-5*s)/s);   // Note the ""+getValue() is used to convert the getValue to a String
    text(lable,x/s,(y+h/2+border_y+13*s)/s);
    popMatrix();  // matrix was pushed for the scale
  }
  
  boolean  handleClick(float mx, float my)
  {
    if((mx>x-(w/2+border_x) && mx<x+(w/2+border_x) && my>y-(h/2+button_height+border_y) && my<y+(h/2+button_height+border_y)) )
    {
      dragging = true;
      handleMove(mx,my);
      return true;  
    }
    else
    {
      return false;
    }    
  }
  
  
  boolean handleRelease(float mx, float my)
  {
    dragging = false;
    return false;  
  }
  
  boolean handleMove(float mx, float my)
  {
    if (dragging)
    {
      yp=my;
      yp=constrain(yp,y-h/2,y+h/2);  
      value = map(yp,y+h/2,y-h/2,0,1);
      value = constrain(value,0,1);
      return true;
    }
    else
    {
      return false;
    }
  }
  
  


  float getValue()
  {
    float v = start + (end-start)*value;
    // scale to nearest resolution
    v = v - (v % resolution);
    return v;
  }
}


class MUICheckBox extends MUIElement
{

  MUICheckBox(float _x, float _y, float _w, float _h, float _value, String _lable)
  {
    x = _x;
    y = _y;
    w = _w;
    h = _h;
    value = _value;
    lable = _lable;
    
    border_x = 5;
    border_y = 5;
    
  }
  
  void display()
  {
    strokeWeight(2);
    rectMode(CENTER);
    fill(0);
    noStroke();
    rect(x,y,w,h);
    stroke(128);
    noFill();
    rect(x,y,w,h);
    
    strokeWeight(2);
    if (value != 0)
    {
      line(x-w/2,y-h/2,x+w/2,y+h/2);
      line(x-w/2,y+h/2,x+w/2,y-h/2);
    } 

    stroke(0,255,0);
    fill(0,255,0,255);
    text(lable,x+w+5,y+4);


/*    
    if (mousePressed)
    {
      strokeWeight(1);
      line(mouseX-w/2,mouseY-h/2,mouseX+w/2,mouseY+h/2);
      line(mouseX-w/2,mouseY+h/2,mouseX+w/2,mouseY-h/2);
    }
*/    
  }
  
  /*
  void update()
  {
    if(mousePressed && (mouseX>x-(w/2+border_x) && mouseX<x+(w/2+border_x) && mouseY>y-(h/2+border_y) && mouseY<y+(h/2+border_y)) )
    {
      value = (value == 0)?1:0;  
    }
  }
  */
  
  boolean  handleClick(float mx, float my)
  {
    if((mx > x-(w/2+border_x) && mx < x+(w/2+border_x) && my > y-(h/2+border_y) && my < y+(h/2+border_y)) )
    {
      value = (value == 0)?1:0;
      return true;  
    }
    else
    {
      return false;
    }
    
  }
  
  boolean checked()
  {
    return value != 0;
  }
  
}

class MUIDragable extends MUIElement
{
  
  boolean dragging;
  float offset_x, offset_y;

  MUIDragable(float _x, float _y, float _w, float _h, float _value, String _lable)
  {
    x = _x;
    y = _y;
    w = _w;
    h = _h;
    value = _value;
    lable = _lable;
    
    border_x = 5;
    border_y = 5;

    dragging = false;
    
  }
  
  void display()
  {
    strokeWeight(2);
    rectMode(CENTER);
    fill(0);
    noStroke();
    rect(x,y,w,h);
    stroke(128);
    noFill();
    rect(x,y,w,h);
    
    strokeWeight(2);
    if (value != 0)
    {
      line(x-w/2,y-h/2,x+w/2,y+h/2);
      line(x-w/2,y+h/2,x+w/2,y-h/2);
    } 

    stroke(0,255,0);
    fill(0,255,0,255);
    text(lable,x+w+5,y+4);
  }
  
  boolean  handleClick(float mx, float my)
  {
    if((mx > x-(w/2+border_x) && mx < x+(w/2+border_x) && my > y-(h/2+border_y) && my < y+(h/2+border_y)) )
    {
      offset_x = mx-x;
      offset_y = my-y;
      dragging = true;
      return true;  
    }
    else
    {
      return false;
    }    
  }
  
  
  boolean handleRelease(float mx, float my)
  {
    dragging = false;
    return false;  
  }
  
  boolean handleMove(float mx, float my)
  {
    if (dragging)
    {
      x = mx-offset_x;
      y = my-offset_y;
      return true;
    }
    else
    {
      return false;
    }
  }
  
}


// The slider manger class simply mantains a list of the sliders
// and runs the update and display functions on them, taking care of 
// removing any dead sliders.
class MUIElementManager
{
  ArrayList elements;

  boolean mouse_pressed   = false;
  boolean mouse_clicked   = false;
  boolean mouse_released  = false;
  boolean mouse_moved;
  float mouse_x = 0;
  float mouse_y = 0;  

  MUIElementManager()
  {
    elements = new ArrayList();
  }

  

  void update()
  {

    if (mouse_x != mouseX || mouse_y != mouseY)
    {
      mouse_moved = true;
    }

    mouse_x = mouseX;
    mouse_y = mouseY;
    
    if (mousePressed)
    {
      if (mouse_pressed == false)
      {
        mouse_clicked = true;
      }
      mouse_pressed = true;
    }
    else
    {
      if (mouse_pressed)
      {
        mouse_released = true;
      }
      // this is a bit long winded, but I'm doing it to avoid reading mousePressed twice, which could 
      // screw up the logic if it's updated asyn
      mouse_pressed = false;
    }
    
    for (int i = elements.size()-1; i >= 0; i--) {
      MUIElement p = (MUIElement) elements.get(i);
      if (!p.dead())
      {
        if (mouse_clicked)
        {
          if (p.handleClick(mouse_x, mouse_y))
          {
            mouse_clicked = false;
          }
        }
      

        if (mouse_released)
        {
          if (p.handleRelease(mouse_x, mouse_y))
          {
            mouse_released = false;
          }
        }
        

        if (mouse_moved)
        {
          if (p.handleMove(mouse_x, mouse_y))
          {
            mouse_moved = false;
          }
        }
        

        if (!p.dead())  // click handledling might have killed it
        {
          p.update();
        }
      }
    }        
    mouse_clicked = false;  // if handled the click, then drop it    
    mouse_released = false;  // if handled the click, then drop it    
    mouse_moved = false;  // if handled the click, then drop it    
    prune();
  }

  void display()
  {
    for (int i = elements.size()-1; i >= 0; i--) {
      MUIElement p = (MUIElement) elements.get(i);
      if (!p.dead() && p.visible)
      {
        p.display();
      }
    }
  }

  void prune()
  {
    for (int i = elements.size()-1; i >= 0; i--) 
    {
      MUIElement p = (MUIElement) elements.get(i);
      if (p.dead()) {
        elements.remove(i);
      }
    }
  }

// Utility Functions, you could just do 
//   MUIMan.add(new MUIVSlider(...));
// but this lets you do:
//   MUIMan.addVSlider(...)
// which is a bit nicer, and encapsulates things slightly if I need to change the registering method below

  MUIVSlider addVSlider(float _x, float _y, float _w, float _h, float _start, float _end, float _initial, float _resolution, String lable_)
  {
    MUIVSlider s = new MUIVSlider(_x,_y,_w,_h,_start,_end,_initial,_resolution, lable_);
    elements.add(s);
    return s;
  }

  MUICheckBox addCheckBox(float _x, float _y, float _w, float _h, float _value, String _lable)
  {
    MUICheckBox s = new MUICheckBox(_x, _y, _w, _h, _value,_lable);
    elements.add(s);
    return s;
  }

  MUIDragable addDragable(float _x, float _y, float _w, float _h, float _value, String _lable)
  {
    MUIDragable s = new MUIDragable(_x, _y, _w, _h, _value,_lable);
    elements.add(s);
    return s;
  }


  void remove(MUIElement s)
  {
    // Mark it as dead
    s.die();
    prune();
  }
}

